from unittest.mock import patch, MagicMock
from code_provenance.registry import fetch_oci_labels, get_registry_token
from code_provenance.models import ImageRef


class TestGetRegistryToken:
    @patch("code_provenance.registry.requests.get")
    def test_ghcr_token(self, mock_get):
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: {"token": "test-token-123"},
        )
        token = get_registry_token("ghcr.io", "acme-org/excalidraw")
        assert token == "test-token-123"
        mock_get.assert_called_once_with(
            "https://ghcr.io/token",
            params={"scope": "repository:acme-org/excalidraw:pull"},
            timeout=10,
        )

    @patch("code_provenance.registry.requests.get")
    def test_docker_hub_token(self, mock_get):
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: {"token": "hub-token-456"},
        )
        token = get_registry_token("docker.io", "library/postgres")
        assert token == "hub-token-456"
        mock_get.assert_called_once_with(
            "https://auth.docker.io/token",
            params={
                "service": "registry.docker.io",
                "scope": "repository:library/postgres:pull",
            },
            timeout=10,
        )


class TestFetchOciLabels:
    @patch("code_provenance.registry.requests.get")
    @patch("code_provenance.registry.get_registry_token")
    def test_returns_labels_when_present(self, mock_token, mock_get):
        mock_token.return_value = "fake-token"

        manifest_response = MagicMock(
            status_code=200,
            json=lambda: {
                "config": {"digest": "sha256:abc123"},
            },
        )
        config_response = MagicMock(
            status_code=200,
            json=lambda: {
                "config": {
                    "Labels": {
                        "org.opencontainers.image.source": "https://github.com/owner/repo",
                        "org.opencontainers.image.revision": "deadbeef1234",
                    }
                }
            },
        )
        mock_get.side_effect = [manifest_response, config_response]

        ref = ImageRef("ghcr.io", "owner", "repo", "v1.0", "ghcr.io/owner/repo:v1.0")
        labels = fetch_oci_labels(ref)
        assert labels["org.opencontainers.image.source"] == "https://github.com/owner/repo"
        assert labels["org.opencontainers.image.revision"] == "deadbeef1234"

    @patch("code_provenance.registry.requests.get")
    @patch("code_provenance.registry.get_registry_token")
    def test_returns_empty_dict_on_failure(self, mock_token, mock_get):
        mock_token.return_value = "fake-token"
        mock_get.return_value = MagicMock(status_code=404)

        ref = ImageRef("ghcr.io", "owner", "repo", "v1.0", "ghcr.io/owner/repo:v1.0")
        labels = fetch_oci_labels(ref)
        assert labels == {}



import os
import re
from code_provenance.models import ImageRef, ImageResult
from code_provenance.registry import fetch_oci_labels
from code_provenance.github import (
    resolve_tag_to_commit, infer_repo_from_dockerhub,
    resolve_ghcr_digest_via_packages, resolve_ghcr_latest_via_packages,
    get_latest_release_commit, get_latest_commit, get_branch_commit,
    find_ghcr_version_by_tag_prefix,
)

_COMMIT_SHA_RE = re.compile(r"^[0-9a-f]{40,}$")
_DIGEST_RE = re.compile(r"^sha256:[0-9a-f]{64}$")


def _is_resolvable_tag(tag: str) -> bool:
    """Check if a tag can be matched against git tags."""
    return bool(tag) and tag != "latest" and not _DIGEST_RE.match(tag)


def resolve_image(service: str, ref: ImageRef) -> ImageResult:
    """Run the resolution chain for a single image."""
    result = ImageResult(
        service=service,
        image=ref.raw,
        registry=ref.registry,
        tag=ref.tag,
    )

    # Step 1: Check OCI labels
    result.steps.append(f"[1/5] Fetching OCI labels from {ref.registry}/{ref.namespace}/{ref.name}:{ref.tag}")
    labels = fetch_oci_labels(ref)
    source = labels.get("org.opencontainers.image.source")
    revision = labels.get("org.opencontainers.image.revision")
    if source and revision:
        result.steps.append(f"[1/5] Found OCI labels: source={source}, revision={revision[:12]}")
        result.repo = source
        result.commit = revision
        result.commit_url = f"{source}/commit/{revision}"
        result.status = "resolved"
        result.resolution_method = "oci_labels"
        result.confidence = "approximate" if ref.tag == "latest" else "exact"
        return result
    else:
        result.steps.append("[1/5] No OCI labels found")

    # Step 2: Infer repo
    result.steps.append(f"[2/5] Inferring GitHub repo from {ref.registry}/{ref.namespace}/{ref.name}")
    owner, repo_name = _infer_repo(ref)
    if owner and repo_name:
        result.steps.append(f"[2/5] Repo inferred: {owner}/{repo_name}")
        result.repo = f"https://github.com/{owner}/{repo_name}"
    else:
        result.steps.append("[2/5] Could not infer GitHub repo")
        result.status = "repo_not_found"
        return result

    # Check if tag is a commit SHA
    if _COMMIT_SHA_RE.match(ref.tag):
        result.steps.append("[2/5] Tag is a commit SHA, using directly")
        result.commit = ref.tag
        result.commit_url = f"{result.repo}/commit/{ref.tag}"
        result.status = "resolved"
        result.resolution_method = "commit_sha_tag"
        result.confidence = "exact"
        return result

    # Step 3: Tag-to-commit resolution
    if _is_resolvable_tag(ref.tag):
        result.steps.append(f'[3/5] Matching tag "{ref.tag}" against git tags in {owner}/{repo_name}')
        tag_result = resolve_tag_to_commit(owner, repo_name, ref.tag)
        if tag_result:
            commit_sha, is_exact = tag_result
            if is_exact:
                result.steps.append(f"[3/5] Exact tag match: {commit_sha[:12]}")
            else:
                result.steps.append(f"[3/5] Prefix match (e.g. v2.10 -> v2.10.x): {commit_sha[:12]}")
            result.commit = commit_sha
            result.commit_url = f"{result.repo}/commit/{commit_sha}"
            result.status = "resolved"
            result.resolution_method = "tag_match"
            result.confidence = "exact" if is_exact else "approximate"
            return result
        result.steps.append("[3/5] No matching git tag found")
        # Try matching a branch with the same name as the tag
        result.steps.append(f"[3/5] Trying branch '{ref.tag}' in {owner}/{repo_name}")
        branch_sha = get_branch_commit(owner, repo_name, ref.tag)
        if branch_sha:
            result.steps.append(f"[3/5] Branch '{ref.tag}' found: {branch_sha[:12]}")
            result.commit = branch_sha
            result.commit_url = f"{result.repo}/commit/{branch_sha}"
            result.status = "resolved"
            result.resolution_method = "branch_match"
            result.confidence = "approximate"
            return result
        result.steps.append(f"[3/5] No branch '{ref.tag}' found")
        # Try GHCR cross-lookup: search for a matching package version tag
        if ref.registry == "docker.io" and os.environ.get("GITHUB_TOKEN"):
            ghcr_owner = ref.name if ref.namespace == "library" else ref.namespace
            result.steps.append(f"[3/5] Trying GHCR package '{ghcr_owner}/{ref.name}' for tag prefix '{ref.tag}'")
            ghcr_match = find_ghcr_version_by_tag_prefix(ghcr_owner, ref.name, ref.tag)
            if ghcr_match:
                matched_tag = ghcr_match["version_tag"]
                result.repo = f"https://github.com/{ghcr_match['repo']}"
                result.steps.append(f"[3/5] GHCR prefix match: '{matched_tag}' in {ghcr_match['repo']}")
                # Try OCI labels on the matched GHCR image for a commit
                ghcr_ref = ImageRef(
                    registry="ghcr.io", namespace=ghcr_owner, name=ref.name,
                    tag=matched_tag, raw=f"ghcr.io/{ghcr_owner}/{ref.name}:{matched_tag}",
                )
                ghcr_labels = fetch_oci_labels(ghcr_ref)
                revision = ghcr_labels.get("org.opencontainers.image.revision")
                if revision:
                    result.steps.append(f"[3/5] OCI labels on GHCR image: revision={revision[:12]}")
                    result.commit = revision
                    result.commit_url = f"{result.repo}/commit/{revision}"
                else:
                    result.steps.append(f"[3/5] No commit in OCI labels, matched version: {matched_tag}")
                result.matched_tag = matched_tag
                result.status = "resolved"
                result.resolution_method = "ghcr_cross_lookup"
                result.confidence = "approximate"
                return result
            result.steps.append("[3/5] No GHCR package match found")
        result.status = "repo_found_tag_not_matched"
        return result

    # Step 4: For GHCR images, try the packages API for digest or :latest
    if ref.registry == "ghcr.io":
        if not os.environ.get("GITHUB_TOKEN"):
            result.steps.append("[4/5] GITHUB_TOKEN not set — skipping packages API (required for digest/:latest resolution)")
            if _DIGEST_RE.match(ref.tag) or ref.tag == "latest" or not ref.tag:
                result.status = "error_no_token"
                return result
            pkg_result = None
            pkg_confidence = None
        elif _DIGEST_RE.match(ref.tag):
            result.steps.append(f"[4/5] Trying GHCR packages API for digest {ref.tag[:20]}...")
            pkg_result = resolve_ghcr_digest_via_packages(ref.namespace, ref.name, ref.tag)
            pkg_confidence = "exact"  # digest is immutable
        elif ref.tag == "latest" or not ref.tag:
            result.steps.append("[4/5] Trying GHCR packages API for :latest tag")
            pkg_result = resolve_ghcr_latest_via_packages(ref.namespace, ref.name)
            pkg_confidence = "approximate"  # :latest is mutable
        else:
            pkg_result = None
            pkg_confidence = None

        if pkg_result:
            repo_full = pkg_result["repo"]
            result.repo = f"https://github.com/{repo_full}"
            if pkg_result.get("commit"):
                commit = pkg_result["commit"]
                pkg_tags = pkg_result.get("tags", [])
                result.steps.append(f"[4/5] Packages API: repo={repo_full}, commit={commit[:12]}, tags={pkg_tags}")
                result.commit = commit
                result.commit_url = f"{result.repo}/commit/{result.commit}"
                if pkg_tags:
                    result.matched_tag = pkg_tags[0] if len(pkg_tags) == 1 else ", ".join(pkg_tags)
                result.status = "resolved"
                result.resolution_method = "packages_api"
                result.confidence = pkg_confidence
                return result
            result.steps.append("[4/5] Packages API: found package but no resolvable tags")
            tags = pkg_result.get("tags", [])
            resolvable = [t for t in tags if t != "latest"]
            result.status = "repo_found_tag_not_matched" if resolvable else "no_tag"
            return result

    # Step 5: For :latest on any registry, try the latest GitHub release,
    # then fall back to the latest commit on the default branch
    if (ref.tag == "latest" or not ref.tag) and owner and repo_name:
        result.steps.append(f"[5/5] Trying latest GitHub release for {owner}/{repo_name}")
        release_result = get_latest_release_commit(owner, repo_name)
        if release_result:
            commit_sha, tag_name = release_result
            result.steps.append(f"[5/5] Latest release: tag={tag_name}, commit={commit_sha[:12]}")
            result.commit = commit_sha
            result.commit_url = f"{result.repo}/commit/{commit_sha}"
            result.status = "resolved"
            result.resolution_method = "latest_release"
            result.confidence = "approximate"
            return result

        # No releases — fall back to latest commit on default branch
        result.steps.append("[5/5] No release found, trying latest commit on default branch")
        latest_sha = get_latest_commit(owner, repo_name)
        if latest_sha:
            result.steps.append(f"[5/5] Latest commit: {latest_sha[:12]}")
            result.commit = latest_sha
            result.commit_url = f"{result.repo}/commit/{latest_sha}"
            result.status = "resolved"
            result.resolution_method = "latest_commit"
            result.confidence = "approximate"
            return result

    result.steps.append("[5/5] Could not resolve to a commit")
    result.status = "no_tag"
    return result


def _infer_repo(ref: ImageRef) -> tuple[str | None, str | None]:
    """Infer the GitHub owner and repo name from an image reference."""
    if ref.registry == "ghcr.io":
        return ref.namespace, ref.name

    if ref.registry == "docker.io":
        hub_result = infer_repo_from_dockerhub(ref.namespace, ref.name)
        if hub_result:
            return hub_result

    return None, None



import json
from code_provenance.models import ImageResult
from code_provenance.output import format_json, format_text


def _sample_results():
    return [
        ImageResult(
            service="web",
            image="ghcr.io/owner/repo:v1.0",
            registry="ghcr.io",
            repo="https://github.com/owner/repo",
            tag="v1.0",
            commit="abc123def456",
            commit_url="https://github.com/owner/repo/commit/abc123def456",
            status="resolved",
            resolution_method="tag_match",
        ),
        ImageResult(
            service="cache",
            image="redis:7.2",
            registry="docker.io",
            tag="7.2",
            status="repo_not_found",
        ),
    ]


class TestFormatJson:
    def test_output_is_valid_json(self):
        output = format_json(_sample_results())
        data = json.loads(output)
        assert len(data) == 2
        assert data[0]["service"] == "web"
        assert data[0]["commit"] == "abc123def456"
        assert data[1]["status"] == "repo_not_found"
        assert data[1]["commit"] is None


class TestFormatText:
    def test_contains_service_names(self):
        output = format_text(_sample_results())
        assert "web" in output
        assert "cache" in output

    def test_contains_status(self):
        output = format_text(_sample_results())
        assert "resolved" in output
        assert "repo_not_found" in output

    def test_contains_commit_url(self):
        output = format_text(_sample_results())
        assert "github.com/owner/repo/commit/abc123def456" in output



import json
from dataclasses import asdict
from code_provenance.models import ImageResult


def format_json(results: list[ImageResult]) -> str:
    """Format results as a JSON array."""
    return json.dumps([asdict(r) for r in results], indent=2)


def format_text(results: list[ImageResult]) -> str:
    """Format results as plain text, one block per image."""
    blocks = []
    for r in results:
        commit = r.commit[:12] if r.commit else "-"
        repo = r.repo.replace("https://", "") if r.repo else "-"
        confidence = r.confidence or "-"
        lines = [
            f"{r.service}: {r.image}",
            f"  repo:       {repo}",
            f"  commit:     {commit}",
            f"  status:     {r.status}",
            f"  confidence: {confidence}",
        ]
        if r.commit_url:
            lines.append(f"  url:        {r.commit_url}")
        if r.matched_tag and r.resolution_method == "ghcr_cross_lookup":
            lines.append(f"  note:       commit is from matched tag '{r.matched_tag}', not the exact image digest")
        elif r.matched_tag:
            lines.append(f"  tag:        {r.matched_tag}")
        blocks.append("\n".join(lines))
    return "\n\n".join(blocks)



"""Integration tests that hit real registries and GitHub API.

Skip with: pytest -m 'not integration'
"""
import pytest
from code_provenance.compose_parser import parse_image_ref
from code_provenance.resolver import resolve_image


@pytest.mark.integration
def test_dockerhub_tag_resolution():
    """Test against real traefik:v3.6.0 on Docker Hub."""
    ref = parse_image_ref("traefik:v3.6.0")
    result = resolve_image("web", ref)
    assert result.repo == "https://github.com/traefik/traefik"
    assert result.commit is not None
    assert len(result.commit) >= 12
    assert result.status == "resolved"
    assert result.confidence == "exact"


@pytest.mark.integration
def test_commit_sha_tag():
    """Test that a commit-SHA tag is detected directly."""
    sha = "ac99122bcbd69f56a7d6523cbc883df9c4766e4c1046b661b76803087e4f475a"
    ref = parse_image_ref(f"ghcr.io/someorg/somerepo:{sha}")
    result = resolve_image("svc", ref)
    assert result.status == "resolved"
    assert result.resolution_method == "commit_sha_tag"
    assert result.commit == sha



import os
from unittest.mock import patch, MagicMock
from code_provenance.github import resolve_tag_to_commit, infer_repo_from_dockerhub, github_headers, check_github_repo_exists


class TestGithubHeaders:
    @patch.dict(os.environ, {}, clear=True)
    def test_no_token(self):
        h = github_headers()
        assert "Authorization" not in h
        assert h["Accept"] == "application/vnd.github+json"

    @patch.dict(os.environ, {"GITHUB_TOKEN": "ghp_test123"})
    def test_with_token(self):
        h = github_headers()
        assert h["Authorization"] == "Bearer ghp_test123"


class TestResolveTagToCommit:
    @patch("code_provenance.github.requests.get")
    def test_exact_match(self, mock_get):
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: [
                {"name": "v3.4.12", "commit": {"sha": "0f769068b3f1abcdef"}},
                {"name": "v3.4.11", "commit": {"sha": "aaa111bbb222"}},
            ],
            links={},
        )
        sha, is_exact = resolve_tag_to_commit("acme-org", "excalidraw", "v3.4.12")
        assert sha == "0f769068b3f1abcdef"
        assert is_exact is True

    @patch("code_provenance.github.requests.get")
    def test_match_with_v_prefix(self, mock_get):
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: [
                {"name": "v3.4.12", "commit": {"sha": "0f769068b3f1abcdef"}},
            ],
            links={},
        )
        sha, is_exact = resolve_tag_to_commit("acme-org", "excalidraw", "3.4.12")
        assert sha == "0f769068b3f1abcdef"
        assert is_exact is True

    @patch("code_provenance.github.requests.get")
    def test_no_match(self, mock_get):
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: [
                {"name": "v1.0.0", "commit": {"sha": "aaa111"}},
            ],
            links={},
        )
        result = resolve_tag_to_commit("owner", "repo", "v9.9.9")
        assert result is None

    @patch("code_provenance.github.requests.get")
    def test_paginated_tags(self, mock_get):
        page1 = MagicMock(
            status_code=200,
            json=lambda: [{"name": f"v0.{i}", "commit": {"sha": f"sha{i}"}} for i in range(100)],
            links={"next": {"url": "https://api.github.com/repos/o/r/tags?page=2"}},
        )
        page2 = MagicMock(
            status_code=200,
            json=lambda: [{"name": "v1.0.0", "commit": {"sha": "target_sha"}}],
            links={},
        )
        mock_get.side_effect = [page1, page2]
        sha, is_exact = resolve_tag_to_commit("o", "r", "v1.0.0")
        assert sha == "target_sha"
        assert is_exact is True

    @patch("code_provenance.github.requests.get")
    def test_prefix_match_picks_highest_version(self, mock_get):
        """v2.10 should match the highest v2.10.x tag."""
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: [
                {"name": "v2.10.7", "commit": {"sha": "sha_v2107"}},
                {"name": "v2.10.6", "commit": {"sha": "sha_v2106"}},
                {"name": "v2.10.0", "commit": {"sha": "sha_v2100"}},
                {"name": "v2.9.0", "commit": {"sha": "sha_v290"}},
            ],
            links={},
        )
        sha, is_exact = resolve_tag_to_commit("traefik", "traefik", "v2.10")
        assert sha == "sha_v2107"
        assert is_exact is False

    @patch("code_provenance.github.requests.get")
    def test_prefix_match_does_not_match_different_minor(self, mock_get):
        """v2.1 should not match v2.10.x."""
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: [
                {"name": "v2.10.7", "commit": {"sha": "sha_v2107"}},
            ],
            links={},
        )
        result = resolve_tag_to_commit("owner", "repo", "v2.1")
        assert result is None

    @patch("code_provenance.github.requests.get")
    def test_exact_match_preferred_over_prefix(self, mock_get):
        """If exact match exists, don't use prefix match."""
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: [
                {"name": "v2.10", "commit": {"sha": "sha_exact"}},
                {"name": "v2.10.7", "commit": {"sha": "sha_v2107"}},
            ],
            links={},
        )
        sha, is_exact = resolve_tag_to_commit("owner", "repo", "v2.10")
        assert sha == "sha_exact"
        assert is_exact is True


class TestCheckGithubRepoExists:
    @patch("code_provenance.github.requests.get")
    def test_exists(self, mock_get):
        mock_get.return_value = MagicMock(status_code=200)
        assert check_github_repo_exists("traefik", "traefik") is True

    @patch("code_provenance.github.requests.get")
    def test_not_exists(self, mock_get):
        mock_get.return_value = MagicMock(status_code=404)
        assert check_github_repo_exists("nonexistent", "repo") is False


class TestInferRepoFromDockerhub:
    @patch("code_provenance.github.check_github_repo_exists")
    def test_official_image_tries_name_as_org(self, mock_exists):
        """For library/traefik, try traefik/traefik on GitHub first."""
        mock_exists.return_value = True
        owner, repo = infer_repo_from_dockerhub("library", "traefik")
        assert owner == "traefik"
        assert repo == "traefik"
        mock_exists.assert_called_once_with("traefik", "traefik")

    @patch("code_provenance.github.requests.get")
    @patch("code_provenance.github.check_github_repo_exists")
    def test_falls_back_to_description(self, mock_exists, mock_get):
        mock_exists.return_value = False
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: {
                "full_description": "Source: https://github.com/docker-library/postgres",
                "description": "The PostgreSQL object-relational database system",
            },
        )
        owner, repo = infer_repo_from_dockerhub("library", "postgres")
        assert owner == "docker-library"
        assert repo == "postgres"

    @patch("code_provenance.github.requests.get")
    @patch("code_provenance.github.check_github_repo_exists")
    def test_no_github_url(self, mock_exists, mock_get):
        mock_exists.return_value = False
        mock_get.return_value = MagicMock(
            status_code=200,
            json=lambda: {
                "full_description": "Some image with no GitHub link",
                "description": "",
            },
        )
        result = infer_repo_from_dockerhub("someuser", "someimage")
        assert result is None



import argparse
import os
import sys
from pathlib import Path
from code_provenance.compose_parser import parse_compose, parse_image_ref
from code_provenance.resolver import resolve_image
from code_provenance.output import format_json, format_text


def main(argv: list[str] | None = None) -> int:
    parser = argparse.ArgumentParser(
        prog="code-provenance",
        description="Resolve Docker images to their source code commits on GitHub.",
    )
    parser.add_argument(
        "compose_file",
        nargs="?",
        default="docker-compose.yml",
        help="Path to docker-compose file (default: docker-compose.yml)",
    )
    parser.add_argument(
        "--json",
        action="store_true",
        dest="json_output",
        help="Output results as JSON",
    )
    parser.add_argument(
        "--verbose", "-v",
        action="store_true",
        help="Show resolution steps",
    )

    args = parser.parse_args(argv)

    if not os.environ.get("GITHUB_TOKEN"):
        print(
            "Warning: GITHUB_TOKEN is not set. Some resolution methods (digest, :latest) will not work.\n"
            "Set it with: export GITHUB_TOKEN=ghp_your_token_here\n"
            "Create a token at https://github.com/settings/tokens with read:packages scope.\n",
            file=sys.stderr,
        )

    compose_path = Path(args.compose_file)
    if not compose_path.exists():
        print(f"Error: {compose_path} not found", file=sys.stderr)
        return 1

    yaml_content = compose_path.read_text()

    try:
        services = parse_compose(yaml_content)
    except Exception as e:
        print(f"Error: failed to parse {compose_path} — {e}", file=sys.stderr)
        return 1

    if not services:
        print("No services with images found. Is this a valid docker-compose file?", file=sys.stderr)
        return 1

    results = []
    for service_name, image_string in services:
        ref = parse_image_ref(image_string)
        result = resolve_image(service_name, ref)
        results.append(result)

    if args.verbose:
        for result in results:
            print(f"\nResolving {result.image} ...", file=sys.stderr)
            for step in result.steps:
                print(f"  {step}", file=sys.stderr)
            print(f"  → {result.status}" + (f" ({result.resolution_method}, {result.confidence})" if result.status == "resolved" else ""), file=sys.stderr)
        print(file=sys.stderr)

    if args.json_output:
        print(format_json(results))
    else:
        print(format_text(results))

    return 0


if __name__ == "__main__":
    sys.exit(main())



import json
import os
import tempfile
from unittest.mock import patch
from code_provenance.cli import main
from code_provenance.models import ImageResult


SAMPLE_COMPOSE = """\
version: '3'
services:
  web:
    image: ghcr.io/acme-org/excalidraw:v3.4.12
    ports:
      - "80:80"
  db:
    image: postgres:16.2
"""


class TestCli:
    @patch("code_provenance.cli.resolve_image")
    def test_json_output(self, mock_resolve, capsys):
        mock_resolve.return_value = ImageResult(
            service="web",
            image="ghcr.io/acme-org/excalidraw:v3.4.12",
            registry="ghcr.io",
            repo="https://github.com/acme-org/excalidraw",
            tag="v3.4.12",
            commit="0f769068b3f1",
            commit_url="https://github.com/acme-org/excalidraw/commit/0f769068b3f1",
            status="resolved",
            resolution_method="tag_match",
        )

        with tempfile.NamedTemporaryFile(mode="w", suffix=".yml", delete=False) as f:
            f.write(SAMPLE_COMPOSE)
            f.flush()
            try:
                main([f.name, "--json"])
                captured = capsys.readouterr()
                data = json.loads(captured.out)
                assert len(data) == 2
                assert data[0]["status"] == "resolved"
            finally:
                os.unlink(f.name)

    def test_missing_file(self, capsys):
        code = main(["/nonexistent/docker-compose.yml"])
        assert code == 1
        captured = capsys.readouterr()
        assert "not found" in captured.err.lower()



import yaml
from code_provenance.models import ImageRef


def parse_image_ref(image_string: str) -> ImageRef:
    """Parse a Docker image string into an ImageRef."""
    raw = image_string

    # Handle digest references (image@sha256:... or image:tag@sha256:...)
    if "@" in image_string:
        name_part, digest = image_string.split("@", 1)
        # Check if there's a tag before the digest (image:tag@sha256:...)
        last_segment = name_part.split("/")[-1]
        if ":" in last_segment:
            colon_pos = name_part.rfind(":")
            tag = name_part[colon_pos + 1:]
            image_string = name_part[:colon_pos]
        else:
            tag = digest
            image_string = name_part
    elif ":" in image_string.split("/")[-1]:
        colon_pos = image_string.rfind(":")
        tag = image_string[colon_pos + 1:]
        image_string = image_string[:colon_pos]
    else:
        tag = "latest"

    # Determine registry
    parts = image_string.split("/")
    if len(parts) >= 2 and ("." in parts[0] or ":" in parts[0]):
        registry = parts[0]
        remaining = parts[1:]
    else:
        registry = "docker.io"
        remaining = parts

    # Determine namespace and name
    if len(remaining) == 1:
        namespace = "library"
        name = remaining[0]
    elif len(remaining) == 2:
        namespace = remaining[0]
        name = remaining[1]
    else:
        namespace = remaining[0]
        name = "/".join(remaining[1:])

    return ImageRef(
        registry=registry,
        namespace=namespace,
        name=name,
        tag=tag,
        raw=raw,
    )


def parse_compose(yaml_content: str) -> list[tuple[str, str]]:
    """Parse docker-compose YAML and return list of (service_name, image_string)."""
    data = yaml.safe_load(yaml_content)
    services = data.get("services", {}) or {}
    results = []
    for service_name, service_config in services.items():
        if isinstance(service_config, dict) and "image" in service_config:
            results.append((service_name, service_config["image"]))
    return results



import requests
from code_provenance.models import ImageRef


def get_registry_token(registry: str, repo_path: str) -> str | None:
    """Get an anonymous pull token from an OCI registry."""
    if registry == "ghcr.io":
        url = "https://ghcr.io/token"
        params = {"scope": f"repository:{repo_path}:pull"}
    elif registry == "docker.io":
        url = "https://auth.docker.io/token"
        params = {
            "service": "registry.docker.io",
            "scope": f"repository:{repo_path}:pull",
        }
    else:
        return None

    try:
        resp = requests.get(url, params=params, timeout=10)
        resp.raise_for_status()
        return resp.json()["token"]
    except (requests.RequestException, KeyError):
        return None


def _registry_base_url(registry: str) -> str:
    if registry == "docker.io":
        return "https://registry-1.docker.io"
    return f"https://{registry}"


_MANIFEST_ACCEPT = ", ".join([
    "application/vnd.docker.distribution.manifest.v2+json",
    "application/vnd.oci.image.manifest.v1+json",
    "application/vnd.docker.distribution.manifest.list.v2+json",
    "application/vnd.oci.image.index.v1+json",
])

_INDEX_MEDIA_TYPES = {
    "application/vnd.docker.distribution.manifest.list.v2+json",
    "application/vnd.oci.image.index.v1+json",
}


def _resolve_manifest_to_config_digest(
    base_url: str, repo_path: str, reference: str, token: str,
) -> str | None:
    """Resolve a manifest reference to a config blob digest, handling multi-arch indexes."""
    headers = {"Authorization": f"Bearer {token}", "Accept": _MANIFEST_ACCEPT}

    try:
        resp = requests.get(
            f"{base_url}/v2/{repo_path}/manifests/{reference}",
            headers=headers,
            timeout=10,
        )
        if resp.status_code != 200:
            return None
        data = resp.json()
    except (requests.RequestException, KeyError, ValueError):
        return None

    media_type = data.get("mediaType", "")

    # If it's an index/manifest list, pick the first amd64/linux manifest
    if media_type in _INDEX_MEDIA_TYPES:
        manifests = data.get("manifests", [])
        platform_digest = None
        for m in manifests:
            platform = m.get("platform", {})
            # Skip attestation manifests
            if platform.get("os") == "unknown":
                continue
            if platform.get("architecture") == "amd64" and platform.get("os") == "linux":
                platform_digest = m["digest"]
                break
        if not platform_digest and manifests:
            # Fall back to first non-attestation manifest
            for m in manifests:
                if m.get("platform", {}).get("os") != "unknown":
                    platform_digest = m["digest"]
                    break
        if not platform_digest:
            return None
        # Recursively resolve the platform-specific manifest
        return _resolve_manifest_to_config_digest(base_url, repo_path, platform_digest, token)

    # Single manifest — extract config digest
    return data.get("config", {}).get("digest")


def fetch_oci_labels(ref: ImageRef) -> dict[str, str]:
    """Fetch OCI labels from an image's config blob without pulling the image."""
    repo_path = f"{ref.namespace}/{ref.name}"
    token = get_registry_token(ref.registry, repo_path)
    if not token:
        return {}

    base_url = _registry_base_url(ref.registry)
    config_digest = _resolve_manifest_to_config_digest(base_url, repo_path, ref.tag, token)
    if not config_digest:
        return {}

    # Fetch config blob to read labels
    try:
        config_headers = {
            "Authorization": f"Bearer {token}",
            "Accept": "application/vnd.docker.container.image.v1+json",
        }
        config_resp = requests.get(
            f"{base_url}/v2/{repo_path}/blobs/{config_digest}",
            headers=config_headers,
            timeout=10,
            allow_redirects=True,
        )
        if config_resp.status_code != 200:
            return {}
        return config_resp.json().get("config", {}).get("Labels", {}) or {}
    except (requests.RequestException, KeyError):
        return {}



from dataclasses import dataclass, field


@dataclass
class ImageRef:
    """A parsed Docker image reference."""
    registry: str          # e.g. "ghcr.io", "docker.io"
    namespace: str         # e.g. "myorg", "library"
    name: str              # e.g. "excalidraw", "postgres"
    tag: str               # e.g. "v3.4.12", "latest"
    raw: str               # original string from docker-compose

    @property
    def full_name(self) -> str:
        """Registry/namespace/name without tag."""
        return f"{self.registry}/{self.namespace}/{self.name}"


@dataclass
class ImageResult:
    """Resolution result for a single image."""
    service: str
    image: str             # original image string
    registry: str
    repo: str | None = None
    tag: str = ""
    commit: str | None = None
    commit_url: str | None = None
    status: str = "repo_not_found"
    resolution_method: str | None = None
    confidence: str | None = None  # "exact", "approximate", or None if unresolved
    matched_tag: str | None = None  # the actual tag matched (when different from the requested tag)
    steps: list[str] = field(default_factory=list)



from code_provenance.compose_parser import parse_compose, parse_image_ref


class TestParseImageRef:
    def test_ghcr_with_tag(self):
        ref = parse_image_ref("ghcr.io/acme-org/excalidraw:v3.4.12")
        assert ref.registry == "ghcr.io"
        assert ref.namespace == "acme-org"
        assert ref.name == "excalidraw"
        assert ref.tag == "v3.4.12"

    def test_docker_hub_official(self):
        ref = parse_image_ref("postgres:16.2")
        assert ref.registry == "docker.io"
        assert ref.namespace == "library"
        assert ref.name == "postgres"
        assert ref.tag == "16.2"

    def test_docker_hub_namespaced(self):
        ref = parse_image_ref("bitnami/redis:7.2")
        assert ref.registry == "docker.io"
        assert ref.namespace == "bitnami"
        assert ref.name == "redis"
        assert ref.tag == "7.2"

    def test_no_tag_defaults_to_latest(self):
        ref = parse_image_ref("nginx")
        assert ref.registry == "docker.io"
        assert ref.namespace == "library"
        assert ref.name == "nginx"
        assert ref.tag == "latest"

    def test_ghcr_no_tag(self):
        ref = parse_image_ref("ghcr.io/owner/repo")
        assert ref.tag == "latest"

    def test_digest_reference(self):
        ref = parse_image_ref("nginx@sha256:abc123")
        assert ref.registry == "docker.io"
        assert ref.namespace == "library"
        assert ref.name == "nginx"
        assert ref.tag == "sha256:abc123"


class TestParseCompose:
    def test_extracts_services_and_images(self):
        yaml_content = """
version: '3'
services:
  web:
    image: ghcr.io/acme-org/excalidraw:v3.4.12
    ports:
      - "80:80"
  db:
    image: postgres:16.2
    environment:
      POSTGRES_PASSWORD: secret
  worker:
    build: ./worker
"""
        services = parse_compose(yaml_content)
        assert len(services) == 2
        assert services[0] == ("web", "ghcr.io/acme-org/excalidraw:v3.4.12")
        assert services[1] == ("db", "postgres:16.2")

    def test_empty_services(self):
        yaml_content = """
version: '3'
services: {}
"""
        services = parse_compose(yaml_content)
        assert services == []



import os
from unittest.mock import patch
from code_provenance.models import ImageRef
from code_provenance.resolver import resolve_image

# Ensure GITHUB_TOKEN is set for tests that need the packages API path
_WITH_TOKEN = patch.dict(os.environ, {"GITHUB_TOKEN": "fake-token-for-tests"})


class TestResolveImage:
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_oci_labels_resolution(self, mock_labels):
        mock_labels.return_value = {
            "org.opencontainers.image.source": "https://github.com/owner/repo",
            "org.opencontainers.image.revision": "abc123def456",
        }
        ref = ImageRef("ghcr.io", "owner", "repo", "v1.0", "ghcr.io/owner/repo:v1.0")
        result = resolve_image("web", ref)
        assert result.status == "resolved"
        assert result.commit == "abc123def456"
        assert result.repo == "https://github.com/owner/repo"
        assert result.resolution_method == "oci_labels"
        assert result.confidence == "exact"

    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_oci_labels_latest_is_approximate(self, mock_labels):
        mock_labels.return_value = {
            "org.opencontainers.image.source": "https://github.com/owner/repo",
            "org.opencontainers.image.revision": "abc123def456",
        }
        ref = ImageRef("ghcr.io", "owner", "repo", "latest", "ghcr.io/owner/repo:latest")
        result = resolve_image("web", ref)
        assert result.status == "resolved"
        assert result.resolution_method == "oci_labels"
        assert result.confidence == "approximate"

    @patch("code_provenance.resolver.resolve_tag_to_commit")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_ghcr_tag_match_fallback(self, mock_labels, mock_tag):
        mock_labels.return_value = {}
        mock_tag.return_value = ("0f769068b3f1", True)

        ref = ImageRef("ghcr.io", "acme-org", "excalidraw", "v3.4.12", "ghcr.io/acme-org/excalidraw:v3.4.12")
        result = resolve_image("web", ref)
        assert result.status == "resolved"
        assert result.commit == "0f769068b3f1"
        assert result.repo == "https://github.com/acme-org/excalidraw"
        assert result.resolution_method == "tag_match"
        assert result.confidence == "exact"
        mock_tag.assert_called_once_with("acme-org", "excalidraw", "v3.4.12")

    @patch("code_provenance.resolver.resolve_tag_to_commit")
    @patch("code_provenance.resolver.infer_repo_from_dockerhub")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_dockerhub_inference_and_tag_match(self, mock_labels, mock_infer, mock_tag):
        mock_labels.return_value = {}
        mock_infer.return_value = ("docker-library", "postgres")
        mock_tag.return_value = ("a1b2c3d4", True)

        ref = ImageRef("docker.io", "library", "postgres", "16.2", "postgres:16.2")
        result = resolve_image("db", ref)
        assert result.status == "resolved"
        assert result.commit == "a1b2c3d4"
        assert result.repo == "https://github.com/docker-library/postgres"
        assert result.confidence == "exact"

    @patch("code_provenance.resolver.resolve_tag_to_commit")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_tag_not_matched(self, mock_labels, mock_tag):
        mock_labels.return_value = {}
        mock_tag.return_value = None

        ref = ImageRef("ghcr.io", "owner", "repo", "v9.9.9", "ghcr.io/owner/repo:v9.9.9")
        result = resolve_image("svc", ref)
        assert result.status == "repo_found_tag_not_matched"
        assert result.commit is None

    def test_commit_sha_as_tag(self):
        sha = "ac99122bcbd69f56a7d6523cbc883df9c4766e4c1046b661b76803087e4f475a"
        ref = ImageRef("ghcr.io", "owner", "repo", sha, f"ghcr.io/owner/repo:{sha}")
        with patch("code_provenance.resolver.fetch_oci_labels", return_value={}):
            result = resolve_image("svc", ref)
        assert result.status == "resolved"
        assert result.commit == sha
        assert result.resolution_method == "commit_sha_tag"
        assert result.confidence == "exact"

    @_WITH_TOKEN
    @patch("code_provenance.resolver.resolve_ghcr_digest_via_packages")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_digest_ref_no_packages_api(self, mock_labels, mock_packages):
        """Digest ref with no packages API result reports no_tag."""
        mock_labels.return_value = {}
        mock_packages.return_value = None
        digest = "sha256:1de50018f208c9a93a21d40b1a830670d113d272b53a45d322c50de9b4db1239"
        ref = ImageRef("ghcr.io", "morpheusais", "morpheus-lumerin-node-tee", digest,
                        f"ghcr.io/morpheusais/morpheus-lumerin-node-tee@{digest}")
        result = resolve_image("proxy-router", ref)
        assert result.status == "no_tag"

    @_WITH_TOKEN
    @patch("code_provenance.resolver.resolve_ghcr_digest_via_packages")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_digest_ref_resolved_via_packages_api(self, mock_labels, mock_packages):
        """Digest ref resolved via packages API."""
        mock_labels.return_value = {}
        mock_packages.return_value = {
            "repo": "MorpheusAIs/Morpheus-Lumerin-Node",
            "commit": "abc123def456",
            "tags": ["v5.0.0"],
        }
        digest = "sha256:1de50018f208c9a93a21d40b1a830670d113d272b53a45d322c50de9b4db1239"
        ref = ImageRef("ghcr.io", "morpheusais", "morpheus-lumerin-node-tee", digest,
                        f"ghcr.io/morpheusais/morpheus-lumerin-node-tee@{digest}")
        result = resolve_image("proxy-router", ref)
        assert result.status == "resolved"
        assert result.commit == "abc123def456"
        assert result.repo == "https://github.com/MorpheusAIs/Morpheus-Lumerin-Node"
        assert result.resolution_method == "packages_api"
        assert result.confidence == "exact"

    @_WITH_TOKEN
    @patch("code_provenance.resolver.resolve_ghcr_digest_via_packages")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_digest_ref_packages_api_no_commit(self, mock_labels, mock_packages):
        """Packages API found the version but no resolvable tags."""
        mock_labels.return_value = {}
        mock_packages.return_value = {
            "repo": "MorpheusAIs/Morpheus-Lumerin-Node",
            "commit": None,
            "tags": [],
        }
        digest = "sha256:1de50018f208c9a93a21d40b1a830670d113d272b53a45d322c50de9b4db1239"
        ref = ImageRef("ghcr.io", "morpheusais", "morpheus-lumerin-node-tee", digest,
                        f"ghcr.io/morpheusais/morpheus-lumerin-node-tee@{digest}")
        result = resolve_image("proxy-router", ref)
        assert result.status == "no_tag"
        assert result.repo == "https://github.com/MorpheusAIs/Morpheus-Lumerin-Node"

    @_WITH_TOKEN
    @patch("code_provenance.resolver.resolve_ghcr_latest_via_packages")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_latest_tag_resolved_via_packages_api(self, mock_labels, mock_latest):
        """GHCR :latest tag resolved via packages API co-tags."""
        mock_labels.return_value = {}
        mock_latest.return_value = {
            "repo": "acme-org/excalidraw",
            "commit": "9891ec84f790",
            "tags": ["v3.24.0", "latest"],
        }
        ref = ImageRef("ghcr.io", "acme-org", "excalidraw", "latest",
                        "ghcr.io/acme-org/excalidraw:latest")
        result = resolve_image("web", ref)
        assert result.status == "resolved"
        assert result.commit == "9891ec84f790"
        assert result.resolution_method == "packages_api"
        assert result.confidence == "approximate"

    @_WITH_TOKEN
    @patch("code_provenance.resolver.get_latest_commit")
    @patch("code_provenance.resolver.get_latest_release_commit")
    @patch("code_provenance.resolver.resolve_ghcr_latest_via_packages")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_latest_tag_no_fallbacks(self, mock_labels, mock_latest, mock_release, mock_commit):
        """GHCR :latest with all fallbacks exhausted reports no_tag."""
        mock_labels.return_value = {}
        mock_latest.return_value = None
        mock_release.return_value = None
        mock_commit.return_value = None
        ref = ImageRef("ghcr.io", "acme-org", "excalidraw", "latest",
                        "ghcr.io/acme-org/excalidraw:latest")
        result = resolve_image("web", ref)
        assert result.status == "no_tag"

    @_WITH_TOKEN
    @patch("code_provenance.resolver.get_latest_commit")
    @patch("code_provenance.resolver.get_latest_release_commit")
    @patch("code_provenance.resolver.resolve_ghcr_latest_via_packages")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_latest_falls_back_to_latest_commit(self, mock_labels, mock_latest, mock_release, mock_commit):
        """When packages API and releases fail, fall back to latest commit."""
        mock_labels.return_value = {}
        mock_latest.return_value = None
        mock_release.return_value = None
        mock_commit.return_value = "abc123def456"
        ref = ImageRef("ghcr.io", "acme-org", "excalidraw", "latest",
                        "ghcr.io/acme-org/excalidraw:latest")
        result = resolve_image("web", ref)
        assert result.status == "resolved"
        assert result.commit == "abc123def456"
        assert result.resolution_method == "latest_commit"
        assert result.confidence == "approximate"

    @patch("code_provenance.resolver.get_latest_release_commit")
    @patch("code_provenance.resolver.infer_repo_from_dockerhub")
    @patch("code_provenance.resolver.fetch_oci_labels")
    def test_dockerhub_latest_resolved_via_latest_release(self, mock_labels, mock_infer, mock_release):
        """Docker Hub :latest resolved via GitHub latest release."""
        mock_labels.return_value = {}
        mock_infer.return_value = ("ollama", "ollama")
        mock_release.return_value = ("abc123def456", "v0.6.2")
        ref = ImageRef("docker.io", "ollama", "ollama", "latest", "ollama/ollama:latest")
        result = resolve_image("ollama", ref)
        assert result.status == "resolved"
        assert result.commit == "abc123def456"
        assert result.resolution_method == "latest_release"
        assert result.confidence == "approximate"



import os
import re
import requests


def github_headers() -> dict[str, str]:
    """Build GitHub API headers, with optional token auth."""
    headers = {"Accept": "application/vnd.github+json"}
    token = os.environ.get("GITHUB_TOKEN")
    if token:
        headers["Authorization"] = f"Bearer {token}"
    return headers


def _normalize_tag(tag: str) -> str:
    """Strip leading 'v' for comparison."""
    return tag.lstrip("v")


def _is_prefix_match(image_tag: str, git_tag: str) -> bool:
    """Check if git_tag is a more specific version of image_tag.

    e.g., image_tag='v2.10' matches git_tag='v2.10.7' but not 'v2.1' or 'v2.100'.
    """
    norm_image = _normalize_tag(image_tag)
    norm_git = _normalize_tag(git_tag)
    return norm_git.startswith(norm_image + ".")


def _parse_version_tuple(tag: str) -> tuple[int, ...] | None:
    """Parse a version string into a tuple of ints for comparison."""
    norm = _normalize_tag(tag)
    # Strip pre-release suffixes like -rc1, -beta2
    norm = re.split(r"[-+]", norm)[0]
    parts = norm.split(".")
    try:
        return tuple(int(p) for p in parts)
    except ValueError:
        return None


def resolve_tag_to_commit(owner: str, repo: str, tag: str) -> tuple[str, bool] | None:
    """Resolve an image tag to a commit SHA by matching against git tags.

    Tries exact match first, then prefix match (e.g., v2.10 -> highest v2.10.x).
    Returns (commit_sha, is_exact_match) or None.
    """
    headers = github_headers()
    url = f"https://api.github.com/repos/{owner}/{repo}/tags"

    prefix_candidates: list[tuple[tuple[int, ...], str]] = []

    while url:
        resp = requests.get(url, headers=headers, params={"per_page": 100}, timeout=10)
        if resp.status_code != 200:
            return None

        for git_tag in resp.json():
            name = git_tag["name"]
            # Exact match (with/without v prefix)
            if name == tag or name == f"v{tag}" or _normalize_tag(name) == _normalize_tag(tag):
                return git_tag["commit"]["sha"], True

            # Collect prefix match candidates
            if _is_prefix_match(tag, name):
                version = _parse_version_tuple(name)
                if version is not None:
                    prefix_candidates.append((version, git_tag["commit"]["sha"]))

        url = resp.links.get("next", {}).get("url")

    # Return the highest version among prefix matches
    if prefix_candidates:
        prefix_candidates.sort(reverse=True)
        return prefix_candidates[0][1], False

    return None


def get_branch_commit(owner: str, repo: str, branch: str) -> str | None:
    """Get the latest commit on a specific branch. Returns SHA or None."""
    headers = github_headers()
    try:
        resp = requests.get(
            f"https://api.github.com/repos/{owner}/{repo}/branches/{branch}",
            headers=headers,
            timeout=10,
        )
        if resp.status_code != 200:
            return None
        return resp.json().get("commit", {}).get("sha")
    except (requests.RequestException, KeyError):
        return None


def get_latest_release_commit(owner: str, repo: str) -> tuple[str, str] | None:
    """Get the commit SHA of the latest GitHub release.

    Returns (commit_sha, tag_name) or None.
    """
    headers = github_headers()
    try:
        resp = requests.get(
            f"https://api.github.com/repos/{owner}/{repo}/releases/latest",
            headers=headers,
            timeout=10,
        )
        if resp.status_code != 200:
            return None
        tag_name = resp.json().get("tag_name")
        if not tag_name:
            return None
    except requests.RequestException:
        return None

    # Resolve the release tag to a commit
    tag_result = resolve_tag_to_commit(owner, repo, tag_name)
    if tag_result:
        commit_sha, _ = tag_result
        return commit_sha, tag_name
    return None


def get_latest_commit(owner: str, repo: str) -> str | None:
    """Get the latest commit SHA on the default branch."""
    headers = github_headers()
    try:
        resp = requests.get(
            f"https://api.github.com/repos/{owner}/{repo}/commits",
            headers=headers,
            params={"per_page": 1},
            timeout=10,
        )
        if resp.status_code != 200:
            return None
        commits = resp.json()
        if commits:
            return commits[0]["sha"]
    except (requests.RequestException, KeyError, IndexError):
        pass
    return None


def check_github_repo_exists(owner: str, repo: str) -> bool:
    """Check if a GitHub repo exists."""
    headers = github_headers()
    try:
        resp = requests.get(
            f"https://api.github.com/repos/{owner}/{repo}",
            headers=headers,
            timeout=10,
        )
        return resp.status_code == 200
    except requests.RequestException:
        return False


def _find_ghcr_package_version(
    owner: str, package_name: str, *, match_digest: str | None = None, match_tag: str | None = None,
) -> dict | None:
    """Find a GHCR package version by digest or tag via the GitHub Packages API.

    Requires GITHUB_TOKEN with read:packages scope.
    Returns {"repo": "owner/repo", "commit": "sha", "tags": [...]} or None.
    """
    headers = github_headers()
    if "Authorization" not in headers:
        return None

    for entity_type in ["orgs", "users"]:
        pkg_base = f"https://api.github.com/{entity_type}/{owner}/packages/container/{package_name}"

        # Get package metadata for source repo
        try:
            pkg_resp = requests.get(pkg_base, headers=headers, timeout=10)
            if pkg_resp.status_code == 403:
                return None
            if pkg_resp.status_code != 200:
                continue
            pkg_data = pkg_resp.json()
        except requests.RequestException:
            continue

        repo_info = pkg_data.get("repository", {})
        full_name = repo_info.get("full_name")
        if not full_name:
            continue

        # Search versions
        url = f"{pkg_base}/versions"
        try:
            while url:
                resp = requests.get(url, headers=headers, params={"per_page": 50}, timeout=10)
                if resp.status_code != 200:
                    break

                for version in resp.json():
                    name = version.get("name", "")
                    metadata = version.get("metadata", {}).get("container", {})
                    tags = metadata.get("tags", [])

                    # Match by digest (version name is the digest)
                    if match_digest and name != match_digest:
                        if match_tag is None:
                            continue
                    # Match by tag
                    if match_tag and match_tag not in tags:
                        continue

                    # Found matching version — resolve tags to a commit
                    repo_owner, repo_name = full_name.split("/", 1)
                    resolvable_tags = [t for t in tags if t != "latest"]
                    for tag in resolvable_tags:
                        tag_result = resolve_tag_to_commit(repo_owner, repo_name, tag)
                        if tag_result:
                            commit_sha, _ = tag_result
                            return {"repo": full_name, "commit": commit_sha, "tags": tags}

                    return {"repo": full_name, "commit": None, "tags": tags}

                url = resp.links.get("next", {}).get("url")
        except requests.RequestException:
            continue

    return None


def resolve_ghcr_digest_via_packages(owner: str, package_name: str, digest: str) -> dict | None:
    """Find the commit for a GHCR image by its digest."""
    return _find_ghcr_package_version(owner, package_name, match_digest=digest)


def resolve_ghcr_latest_via_packages(owner: str, package_name: str) -> dict | None:
    """Find the commit for a GHCR image's :latest tag."""
    return _find_ghcr_package_version(owner, package_name, match_tag="latest")


def find_ghcr_version_by_tag_prefix(
    owner: str, package_name: str, tag_prefix: str,
) -> dict | None:
    """Search GHCR package versions for a tag matching a prefix.

    e.g., tag_prefix='10.11' matches '10.11.16-jammy', returns the first match.
    Returns {"repo": "owner/repo", "version_tag": "10.11.16-jammy"} or None.
    """
    headers = github_headers()
    if "Authorization" not in headers:
        return None

    for entity_type in ["orgs", "users"]:
        pkg_base = f"https://api.github.com/{entity_type}/{owner}/packages/container/{package_name}"

        try:
            pkg_resp = requests.get(pkg_base, headers=headers, timeout=10)
            if pkg_resp.status_code == 403:
                return None
            if pkg_resp.status_code != 200:
                continue
            pkg_data = pkg_resp.json()
        except requests.RequestException:
            continue

        repo_info = pkg_data.get("repository", {})
        full_name = repo_info.get("full_name")
        if not full_name:
            continue

        # Search versions for a tag matching the prefix
        url = f"{pkg_base}/versions"
        best_match = None
        best_version: tuple[int, ...] | None = None
        try:
            while url:
                resp = requests.get(url, headers=headers, params={"per_page": 100}, timeout=10)
                if resp.status_code != 200:
                    break

                for version in resp.json():
                    metadata = version.get("metadata", {}).get("container", {})
                    tags = metadata.get("tags", [])
                    for tag in tags:
                        # Check if tag starts with prefix followed by . or -
                        stripped = tag[len(tag_prefix):] if tag.startswith(tag_prefix) else None
                        if stripped is not None and (stripped == "" or stripped[0] in ".-"):
                            # Parse version for comparison
                            ver_str = tag.split("-")[0]  # strip OS suffix like -jammy
                            parts = ver_str.split(".")
                            try:
                                ver_tuple = tuple(int(p) for p in parts)
                            except ValueError:
                                ver_tuple = None
                            if ver_tuple and (best_version is None or ver_tuple > best_version):
                                best_version = ver_tuple
                                best_match = {"repo": full_name, "version_tag": tag}

                url = resp.links.get("next", {}).get("url")
        except requests.RequestException:
            pass

        if best_match:
            return best_match

    return None


def infer_repo_from_dockerhub(namespace: str, name: str) -> tuple[str, str] | None:
    """Try to find the GitHub repo for a Docker Hub image."""
    # For official images (library/X), try the image name as org/repo directly
    # e.g., traefik -> traefik/traefik, nginx -> nginx/nginx
    if namespace == "library":
        if check_github_repo_exists(name, name):
            return name, name

    # For namespaced images, try namespace/name on GitHub
    if namespace != "library":
        if check_github_repo_exists(namespace, name):
            return namespace, name

    # Fall back to scraping Docker Hub description for GitHub links
    url = f"https://hub.docker.com/v2/repositories/{namespace}/{name}"
    try:
        resp = requests.get(url, timeout=10)
        if resp.status_code != 200:
            return None

        data = resp.json()
        text = (data.get("full_description") or "") + " " + (data.get("description") or "")
        match = re.search(r"https?://github\.com/([\w.-]+)/([\w.-]+)", text)
        if match:
            return match.group(1), match.group(2)
    except requests.RequestException:
        pass

    return None



from code_provenance.models import ImageRef, ImageResult


def test_image_ref_full_name():
    ref = ImageRef(
        registry="ghcr.io",
        namespace="acme-org",
        name="excalidraw",
        tag="v3.4.12",
        raw="ghcr.io/acme-org/excalidraw:v3.4.12",
    )
    assert ref.full_name == "ghcr.io/acme-org/excalidraw"


def test_image_result_defaults():
    r = ImageResult(service="web", image="nginx:latest", registry="docker.io")
    assert r.status == "repo_not_found"
    assert r.commit is None
    assert r.resolution_method is None
